from npiai.utils.logger import logger
from npiai.utils.to_async_fn import to_async_fn
from npiai.utils.sanitize_schema import sanitize_schema
from npiai.utils.parse_docstring import parse_docstring

__all__ = ['logger', 'to_async_fn', 'sanitize_schema', 'parse_docstring']



from npiai import BrowserApp, function, NavigatorAgent
from npiai import LLM

__SYSTEM_PROMPT__ = """
You are a general browser-based autonomous agent helping user to finish any task on any webpage. For a given task, you should first go to the appropriate web page, and then pass the task to the navigator to fulfill it.
"""


class Browser(BrowserApp):
    def __init__(self, navigator_llm: LLM = None, headless: bool = True):
        super().__init__(
            name='browser',
            description='Perform any task on any webpage',
            system_prompt=__SYSTEM_PROMPT__,
            headless=headless,
        )

        self.add(
            NavigatorAgent(
                llm=navigator_llm,
                playwright=self.playwright,
            )
        )

    @function
    async def goto(self, url: str):
        """
        Open the given URL in the browser.

        Args:
            url: The URL to open.
        """
        await self.playwright.page.goto(url)

        return f'Opened {await self.get_page_url()}, page title: {await self.get_page_title()}'



import asyncio
from npiai import agent_wrapper
from npiai.browser_app import Twitter


async def main():
    async with agent_wrapper(Twitter(headless=False)) as twitter:
        return await twitter.chat(
            'Find and summarize the latest tweet by @AtomSilverman discussing AI agents. You should also include the link to the tweet.'
        )


if __name__ == '__main__':
    asyncio.run(main())



import os
from abc import ABC, abstractmethod
from typing import List

import yaml
from litellm.types.completion import ChatCompletionToolMessageParam
from litellm.types.utils import ChatCompletionMessageToolCall

from npiai.types import FunctionRegistration
from npiai.utils import logger
from npiai.core.hitl import HITL


class Tool(ABC):
    name: str
    description: str
    provider: str
    _hitl: HITL | None = None

    @property
    def hitl(self) -> HITL:
        if self._hitl is None:
            raise AttributeError('HITL handler has not been set')

        return self._hitl

    @staticmethod
    def test():
        return 1

    @abstractmethod
    def unpack_functions(self) -> List[FunctionRegistration]:
        """Export the functions registered in the app"""
        ...

    @abstractmethod
    async def start(self):
        """Start the app"""
        ...

    @abstractmethod
    async def end(self):
        """Stop and dispose the app"""
        ...

    def use_hitl(self, hitl: HITL):
        self._hitl = hitl

    def export(self, filename: str | None = None):
        """
        Find the wrapped tool functions and export them as yaml
        """
        data = {
            'kind': 'Function',
            'metadata': {
                'name': self.name,
                'description': self.description,
                'provider': self.provider,
            },
            'spec': {
                'runtime': {
                    'language': 'python',
                    'version': '3.11',
                },
                'dependencies': [{
                    'name': 'npiai',
                    'version': '0.1.0',
                }],
                'functions': [t.get_meta() for t in self.unpack_functions()],
            }
        }

        if filename is not None:
            os.makedirs(os.path.dirname(filename), exist_ok=True)
            with open(filename, 'w') as f:
                yaml.dump(data, f)
                logger.info(f'Exported schema to: {filename}')
        else:
            return data

    # context manager
    async def __aenter__(self):
        await self.start()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.end()


class BaseApp(Tool, ABC):
    @abstractmethod
    async def call(
            self,
            tool_calls: List[ChatCompletionMessageToolCall],
    ) -> List[ChatCompletionToolMessageParam]:
        ...


class BaseAgent(Tool, ABC):
    @abstractmethod
    async def chat(self, message: str) -> str:
        ...



import os

from twilio.rest import Client

from npiai import App, function
from npiai.error.auth import UnauthorizedError


class Twilio(App):
    def __init__(self, account_sid: str = None, auth_token: str = None, from_number: str = None):
        super().__init__(
            name='twilio',
            description='Send messages using Twilio, e.g., twilio("send a message to +16507055795")',
            system_prompt='You are a Twilio Agent can send messages via WhatsApp or SMS',
        )

        sid = account_sid or os.environ.get('TWILIO_ACCOUNT_SID', None)
        token = auth_token or os.environ.get('TWILIO_AUTH_TOKEN', None)

        if not sid:
            raise UnauthorizedError('Twilio account sid is missing')
        if not token:
            raise UnauthorizedError('Twilio auth token is missing')

        self.client = Client(sid, token)
        self.from_number = from_number or os.environ.get('TWILIO_FROM_NUMBER', None)

    @function
    async def send_sms_message(self, to: str, body: str) -> str:
        """
        Send a sms message to a phone number.

        Args:
            to: The phone number you want to send the message to.
            body: The body of the message.
        """
        message = self.client.messages.create(
            from_=f'{self.from_number}',
            body=body,
            to=to,
        )

        return f'Message has been successfully sent, sid: {message.sid}'

    @function
    async def send_whatsapp_message(self, to: str, body: str) -> str:
        """
        Send a whatsapp message to a phone number.

        Args:
            to: The phone number you want to send the message to.
            body: The body of the message.
        """
        message = self.client.messages.create(
            from_=f'whatsapp:{self.from_number}',
            body=body,
            to=f'whatsapp:{to}',
        )

        return f'Message has been successfully sent, sid: {message.sid}'



from npiai.types.function_registration import FunctionRegistration, ToolFunction
from npiai.types.shot import Shot
from npiai.types.tool_meta import ToolMeta

__all__ = ['FunctionRegistration', 'ToolFunction', 'Shot', 'ToolMeta']



from typing import Callable, Optional, Awaitable, List, Dict, Any
from dataclasses import dataclass, asdict

from openai.types.chat import ChatCompletionToolParam

from npiai.types.shot import Shot

ToolFunction = Callable[..., Awaitable[str]]


@dataclass(frozen=True)
class FunctionRegistration:
    fn: ToolFunction
    description: str
    name: str
    schema: Optional[Dict[str, Any]] = None
    few_shots: Optional[List[Shot]] = None

    def get_meta(self):
        return {
            'description': self.description,
            'name': self.name,
            'parameters': self.schema,
            'fewShots': [asdict(ex) for ex in self.few_shots] if self.few_shots else None,
        }

    def get_tool_param(self) -> ChatCompletionToolParam:
        tool: ChatCompletionToolParam = {
            'type': 'function',
            'function': {
                'name': self.name,
                'description': self.description,
            }
        }

        if self.schema is not None:
            tool['function']['parameters'] = self.schema

        return tool



import asyncio

from npiai import agent_wrapper
from npiai import GitHub


async def main():
    async with agent_wrapper(GitHub()) as github:
        return await github.chat('Close all pull requests in idiotWu/npi-test')


if __name__ == '__main__':
    asyncio.run(main())



from .app import BrowserApp
from .playwright_context import PlaywrightContext
from npiai.core.agent.navigator_agent import NavigatorAgent

__all__ = ['BrowserApp', 'PlaywrightContext', 'NavigatorAgent']



from abc import ABC, abstractmethod


class HITL(ABC):
    @abstractmethod
    async def confirm(self, app_name: str, message: str) -> bool:
        ...

    @abstractmethod
    async def input(self, app_name: str, message: str) -> str:
        ...



import asyncio
from npiai import agent_wrapper
from npiai.browser_app import Twitter


async def main():
    async with agent_wrapper(Twitter(headless=False)) as twitter:
        return await twitter.chat(
            'Reply to the latest message in the notifications with your opinions. You can write anything you like.',
        )


if __name__ == '__main__':
    asyncio.run(main())



from .discord import Discord
from .github import GitHub
from .google.gmail import Gmail
from .google.calendar import GoogleCalendar
from .slack import Slack
from .twilio import Twilio

__all__ = ["Discord", "GitHub", "Gmail", "GoogleCalendar", "Slack", "Twilio"]



"""The basic interface for NPi Apps"""
import functools
import inspect
import json
import re
from dataclasses import asdict
from typing import Dict, List, Optional, Any

import yaml
# TODO: use llmlite typings
from openai.types.chat import (
    ChatCompletionToolParam,
    ChatCompletionMessageToolCall,
    ChatCompletionToolMessageParam,
)
from pydantic import Field, create_model

from npiai.core.base import BaseApp, Tool
from npiai.core.hitl import HITL

from npiai.types import FunctionRegistration, ToolFunction, Shot, ToolMeta
from npiai.utils import logger, sanitize_schema, parse_docstring, to_async_fn

__NPI_TOOL_ATTR__ = '__NPI_TOOL_ATTR__'


def function(
    tool_fn: ToolFunction = None,
    name: Optional[str] = None,
    description: Optional[str] = None,
    schema: Dict[str, Any] = None,
    few_shots: Optional[List[Shot]] = None,
):
    """
    NPi Tool decorator for functions

    Args:
        tool_fn: Tool function. This value will be set automatically.
        name: Tool name. The tool function name will be used if not given.
        description: Tool description. This value will be inferred from the tool's docstring if not given.
        schema: Tool parameters schema. This value will be inferred from the tool's type hints if not given.
        few_shots: Predefined working examples.

    Returns:
        Wrapped tool function that will be registered on the app class
    """

    def decorator(fn: ToolFunction):
        setattr(
            fn,
            __NPI_TOOL_ATTR__,
            ToolMeta(
                name=name,
                description=description,
                schema=schema,
                few_shots=few_shots,
            )
        )

        @functools.wraps(fn)
        def wrapper(*args, **kwargs):
            return fn(*args, **kwargs)

        return wrapper

    # called as `@function`
    if callable(tool_fn):
        return decorator(tool_fn)

    # called as `@function(...)`
    return decorator


class App(BaseApp):
    """The basic interface for the natural language programming interface"""

    name: str
    description: str
    system_prompt: Optional[str]
    provider: str

    _tools: List[ChatCompletionToolParam]
    _fn_map: Dict[str, FunctionRegistration]
    _sub_tools: List[Tool]

    _started: bool = False

    @property
    def tools(self):
        return self._tools

    @property
    def functions(self):
        return list(self._fn_map.values())

    @property
    def sub_tools(self):
        return self._sub_tools

    def __init__(
        self,
        name: str,
        description: str,
        system_prompt: str = None,
        provider: str = None,
    ):
        super().__init__()
        self.name = name
        self.description = description
        self.system_prompt = system_prompt
        self.provider = provider or 'private'
        self._tools = []
        self._fn_map = {}
        self._sub_tools = []

        self._register_tools()

    def unpack_functions(self) -> List[FunctionRegistration]:
        return list(self._fn_map.values())

    def use_hitl(self, hitl: HITL):
        """
        Attach the given HITL handler to this app and all its sub apps

        Args:
            hitl: HITL handler
        """
        super().use_hitl(hitl)

        for app in self._sub_tools:
            app.use_hitl(hitl)

    async def start(self):
        """Start the app"""
        if not self._started:
            self._started = True
            for app in self._sub_tools:
                await app.start()

    async def end(self):
        """Stop and dispose the app"""
        if self._started:
            self._started = False
            for app in self._sub_tools:
                await app.end()

    def add_tool(
        self,
        *tools: Tool,
    ):
        for tool in tools:
            # share hitl handler
            if tool._hitl is None:
                tool.use_hitl(self._hitl)

            self._sub_tools.append(tool)

            for fn_reg in tool.unpack_functions():
                data = asdict(fn_reg)
                data['name'] = f'{tool.name}__{fn_reg.name}'
                scoped_fn_reg = FunctionRegistration(**data)

                self._add_function(scoped_fn_reg)

    async def debug(self, app_name: str = None, fn_name: str = None, args: Dict[str, Any] = None) -> str:
        if app_name:
            fn_name = f'{app_name}__{fn_name}'
        return await self._exec(fn_name, args)

    async def call(
        self,
        tool_calls: List[ChatCompletionMessageToolCall],
    ) -> List[ChatCompletionToolMessageParam]:
        results: List[ChatCompletionToolMessageParam] = []

        for call in tool_calls:
            fn_name = call.function.name
            args = json.loads(call.function.arguments)
            call_msg = f'[{self.name}]: Calling {fn_name}'

            if args:
                arg_list = ', '.join(f'{k}={json.dumps(v)}' for k, v in args.items())
                call_msg += f'({arg_list})'
            else:
                call_msg += '()'

            logger.info(call_msg)

            res = await self._exec(fn_name, args)

            logger.debug(f'[{self.name}]: function `{fn_name}` returned: {res}')

            results.append(
                ChatCompletionToolMessageParam(
                    role='tool',
                    tool_call_id=call.id,
                    content=res,
                )
            )

        return results

    async def _exec(self, fn_name: str, args: Dict[str, Any] = None) -> str:
        if fn_name not in self._fn_map:
            raise RuntimeError(
                f'[{self.name}]: function `{fn_name}` not found. Available functions: {self._fn_map.keys()}'
            )

        tool = self._fn_map[fn_name]

        try:
            if args is None:
                return await tool.fn()
            return await tool.fn(**args)
        except Exception as e:
            logger.error(e)
            return f'Error: {e}'

    def _register_tools(self):
        """
        Find the wrapped tool functions and register them in this app
        """
        for attr, fn in inspect.getmembers(self, lambda x: inspect.ismethod(x) or inspect.isfunction(x)):
            tool_meta: ToolMeta | None = getattr(fn, __NPI_TOOL_ATTR__, None)

            if not callable(fn) or not tool_meta:
                continue

            params = list(inspect.signature(fn).parameters.values())
            docstr = parse_docstring(inspect.getdoc(fn))
            tool_name = tool_meta.name or fn.__name__
            tool_desc = tool_meta.description or docstr.description

            if not tool_desc:
                raise ValueError(
                    f'Unable to get the description of tool function `{fn}`'
                )

            # parse schema
            tool_schema = tool_meta.schema

            if not tool_schema and len(params) > 0:
                # get parameter descriptions
                param_descriptions = {}
                for p in docstr.params:
                    param_descriptions[p.arg_name] = p.description

                # get parameter field definitions
                param_fields = {}
                for p in params:
                    param_fields[p.name] = (p.annotation, Field(
                        default=p.default if p.default is not inspect.Parameter.empty else ...,
                        description=param_descriptions.get(p.name, ''),
                    ))

                model = create_model(f'{tool_name}_model', **param_fields)
                tool_schema = sanitize_schema(model)

            # parse examples
            tool_few_shots = tool_meta.few_shots
            doc_shots = [m for m in docstr.meta if m.args == ['few_shots']]

            if not tool_few_shots and len(doc_shots) > 0:
                tool_few_shots = []

                for shot in doc_shots:
                    items = re.findall(r'^\s*- ', shot.description, re.MULTILINE)
                    if len(items) == 1:
                        # remove leading '- ' to avoid indentation issues
                        shot_data = yaml.safe_load(re.sub(r'^\s*- ', '', shot.description))
                    else:
                        shot_data = yaml.safe_load(shot.description)

                    if not isinstance(shot_data, list):
                        shot_data = [shot_data]

                    for d in shot_data:
                        tool_few_shots.append(Shot(**d))

            self._add_function(
                FunctionRegistration(
                    # wrap fn in an async wrapper
                    fn=to_async_fn(fn),
                    name=tool_name,
                    description=tool_desc.strip(),
                    schema=tool_schema,
                    few_shots=tool_few_shots,
                )
            )

    def _add_function(self, fn_reg: FunctionRegistration):
        if fn_reg.name in self._fn_map:
            raise Exception(f'Duplicate function: {fn_reg.name}')

        self._fn_map[fn_reg.name] = fn_reg
        self._tools.append(fn_reg.get_tool_param())



import asyncio

from npiai import agent_wrapper
from npiai.app.discord import Discord


async def main():
    async with agent_wrapper(Discord()) as discord:
        return await discord.chat(
            'Send a direct message to Dolphin (user id: 209243186332303362) asking if he is doing well, and wait for his reply.'
        )


if __name__ == '__main__':
    asyncio.run(main())



from typing import Dict, Any

from pydantic import BaseModel


def sanitize_schema(model: BaseModel) -> Dict[str, Any]:
    schema = model.model_json_schema()

    # remove unnecessary title
    schema.pop('title', None)

    for prop in schema.get('properties', {}).values():
        prop.pop('title', None)

        # use a more compact format for optional fields
        if 'anyOf' in prop and len(prop['anyOf']) == 2 and prop['anyOf'][1]['type'] == 'null':
            # copy the first type definition to props
            t = prop['anyOf'][0]
            for k, v in t.items():
                prop[k] = v

            prop.pop('anyOf', None)

            if prop.get('default') is None:
                prop.pop('default', None)

    # remove empty properties
    if len(schema.get('properties', {})) == 0:
        schema.pop('properties', None)

    return schema



from npiai.core.app import App, function
from npiai.core.app.browser import BrowserApp, PlaywrightContext
from npiai.core.agent import agent_wrapper, Agent, BrowserAgent, NavigatorAgent
from npiai.core.hitl import HITL
from npiai.core.base import BaseAgent, BaseApp, Tool

__all__ = [
    'Tool',
    'function',
    'App',
    'BrowserApp',
    'PlaywrightContext',
    'NavigatorAgent',
    'HITL',
    'Agent',
    'BrowserAgent',
    'agent_wrapper',
]



import asyncio
from npiai import agent_wrapper
from npiai.browser_app.browser import Browser


async def main():
    async with agent_wrapper(Browser(headless=False)) as browser:
        return await browser.chat(
            'Book a one-way flight from ATL to LAX on 4/20 using Google Flights.'
        )


if __name__ == '__main__':
    asyncio.run(main())



import asyncio

from npiai import agent_wrapper
from npiai.app.google import Gmail
from google.oauth2.credentials import Credentials


async def main():
    credentials = Credentials.from_authorized_user_file('gmail_token.json')

    async with agent_wrapper(Gmail(credentials)) as gmail:
        return await gmail.chat('Add label "TEST" to the latest email from daofeng.wu@emory.edu')


if __name__ == '__main__':
    asyncio.run(main())



import questionary

from termcolor import colored
from npiai.core import HITL


class ConsoleHandler(HITL):
    async def confirm(self, app_name: str, message: str) -> bool:
        # print(colored(f'[{app_name}]: Confirmation required', color='green', attrs=['bold']))
        print(colored(f'[{app_name}]: {message}', color='magenta', attrs=['bold']))
        return await questionary.confirm('Continue?').ask_async()

    async def input(self, app_name: str, message: str) -> str:
        # print(colored(f'[{app_name}]: Additional information required', color='green', attrs=['bold']))
        print(colored(f'[{app_name}]: {message}', color='magenta', attrs=['bold']))
        return await questionary.text('Type your response:').ask_async()



# pylint: disable=missing-module-docstring
import asyncio

from npiai import agent_wrapper
from npiai.app.google import GoogleCalendar
from google.oauth2.credentials import Credentials


async def main():
    credentials = Credentials.from_authorized_user_file('gc_token.json')

    async with agent_wrapper(GoogleCalendar(credentials)) as gc:
        return await gc.chat(message='get the incoming 10 events')


if __name__ == '__main__':
    asyncio.run(main())



"""The definition of Google Calendar App"""

import datetime
import json

from google.oauth2.credentials import Credentials
from googleapiclient.discovery import build

from npiai import function, App


# https://developers.google.com/calendar/quickstart/python
# API Reference: https://developers.google.com/calendar/api/v3/reference


class GoogleCalendar(App):
    def __init__(self, credentials: Credentials):
        super().__init__(
            name='google_calendar',
            description='Manage events on Google Calendar',
            system_prompt='You are an assistant interacting with Google Calendar API. Your job is the selecting the best function based the tool list.',
        )

        self.service = build(
            'calendar', 'v3', credentials=credentials
        )

    @function
    async def get_user_email(self, message: str) -> str:
        """
        Get the user's email address

        Args:
            message: The message to ask the user for their email address
        """
        return await self.hitl.input(self.name, message)

    @function
    def get_today(self):
        """Get today's date"""
        return datetime.date.today().strftime('%a, %Y-%m-%d')

    @function
    def get_timezone(self):
        """Get the user's timezone"""
        res = self.service.calendars().get(  # pylint: disable=maybe-no-member
            calendarId='primary'
        ).execute()

        return res.get('timeZone')

    @function
    def retrieve_events(self, calendar_id: str = 'primary', time_min: str = None, time_max: str = None) -> str:
        """
        Retrieve events from Google Calendar.

        Args:
            calendar_id: The calendar ID which typically in the form of an email address. The primary calendar will be used if not given.
            time_min: Lower bound (exclusive) for an event's end time to filter by. Must be an RFC3339 timestamp with mandatory time zone offset, for example, 2011-06-03T10:00:00-05:00. If `time_min` is set, `time_max` must be greater than `time_min`.
            time_max: Upper bound (exclusive) for an event's start time to filter by. Must be an RFC3339 timestamp with mandatory time zone offset, for example, 2011-06-05T10:00:00-05:00. If `time_max` is set, `time_min` must be smaller than `time_min`.
        """
        calendar_id = calendar_id
        time_min = time_min
        time_max = time_max
        max_result = 10
        single_events = True
        order_by = 'startTime'
        event_types = "default"

        if time_min is None:
            # 'Z' indicates UTC time
            time_min = datetime.datetime.utcnow().strftime("%Y-%m-%dT%H:%M:%SZ")  # pylint: disable=maybe-no-member

        events_result = (
            self.service.events().list(  # pylint: disable=maybe-no-member
                calendarId=calendar_id,
                timeMin=time_min,
                timeMax=time_max,
                maxResults=max_result,
                singleEvents=single_events,
                orderBy=order_by,
                eventTypes=event_types,
            ).execute()
        )
        events = events_result.get("items", [])

        if not events:
            return "No events found."

        return json.dumps(events, ensure_ascii=False)

    @function
    def create_event(
        self,
        summary: str,
        start_time: str,
        end_time: str,
        calendar_id: str = 'primary',
        description: str = '',
    ) -> str:
        """
        Create and add an event to Google Calendar.

        Args:
            summary: The summary of the event.
            start_time: The start time of this event. Must be an RFC3339 timestamp with mandatory time zone offset, for example, 2011-06-03T09:00:00-05:00. You should retrieve the user's timezone first.
            end_time: The end time of this event. Must be an RFC3339 timestamp with mandatory time zone offset, for example, 2011-06-03T10:00:00-05:00. You should retrieve the user's timezone first.
            calendar_id: The calendar ID which typically in the form of an email address. The primary calendar will be used if not given.
            description: The detailed description of the event.
        """
        attendee = [],
        location = None,
        recurrence = None,
        event = {
            'summary': summary,
            'location': location,
            'description': description,
            'start': {
                'dateTime': start_time,
            },
            'end': {
                'dateTime': end_time,
            },
            'recurrence': [recurrence],
            'attendees': attendee,
            'reminders': {
                'useDefault': False,
                'overrides': [
                    {'method': 'email', 'minutes': 24 * 60},
                    {'method': 'popup', 'minutes': 10},
                ],
            },
        }

        event = self.service.events().insert(  # pylint: disable=maybe-no-member
            calendarId=calendar_id, body=event
        ).execute()

        return f'Event created: {event.get("htmlLink")}'



import asyncio

from npiai import agent_wrapper
from npiai.app.google import Gmail
from google.oauth2.credentials import Credentials


async def main():
    credentials = Credentials.from_authorized_user_file('gmail_token.json')

    async with agent_wrapper(Gmail(credentials)) as gmail:
        return await gmail.chat(
            'Send an email to daofeng@npi.ai inviting him to join an AI meetup and wait for his reply. The date candidates are: Monday, Tuesday, Wednesday'
        )


if __name__ == '__main__':
    asyncio.run(main())



import pathlib
import tempfile
import os
from urllib.request import urlretrieve

from playwright.async_api import async_playwright, Playwright, Browser, BrowserContext, Page, FileChooser

__BROWSER_UTILS_VERSION__ = '0.0.2'


def _prepare_browser_utils():
    # path to the js bundle
    cache_dir = pathlib.Path(tempfile.gettempdir()) / '.npi'
    js_path = cache_dir / f'browser-utils@{__BROWSER_UTILS_VERSION__}.js'

    if js_path.exists():
        return js_path

    os.makedirs(cache_dir, exist_ok=True)

    urlretrieve(f'https://unpkg.com/@npi-ai/browser-utils@{__BROWSER_UTILS_VERSION__}/dist/index.global.js', js_path)

    return js_path


class PlaywrightContext:
    playwright: Playwright
    browser: Browser
    context: BrowserContext
    page: Page
    headless: bool
    ready: bool

    def __init__(
        self,
        headless: bool = True,
    ):
        """
        Initialize a Playwright context

        Args:
            headless: Whether to run playwright in headless mode
        """
        self.headless = headless
        self.ready = False

    async def start(self):
        """Start the Playwright browser"""
        self.playwright = await async_playwright().start()
        self.browser = await self.playwright.chromium.launch(
            headless=self.headless,
            args=['--disable-gpu', '--single-process'],
        )

        self.context = await self.browser.new_context(
            locale='en-US',
            bypass_csp=True,
            **self.playwright.devices['Desktop Edge'],
        )
        # self.context.set_default_timeout(3000)
        await self.context.add_init_script(path=_prepare_browser_utils())
        await self.context.add_init_script(script="""window.npi = new window.BrowserUtils()""")

        self.page = await self.context.new_page()
        self.page.on('filechooser', self.on_filechooser)
        self.page.on('popup', self.on_popup)

        self.ready = True

    async def on_filechooser(self, chooser: FileChooser):
        """
        Callback function invoked when an input:file is clicked

        Args:
            chooser: FileChooser instance
        """
        await chooser.set_files('')

    async def on_popup(self, popup: Page):
        """
        Callback function invoked when a tab is opened

        Args:
            popup: Page instance
        """
        self.page = popup

    async def stop(self):
        """
        Dispose the browser app
        """
        await self.playwright.stop()



import asyncio

from npiai import agent_wrapper
from npiai.app.slack import Slack


async def main():
    async with agent_wrapper(Slack()) as slack:
        return await slack.chat(
            'Send a direct message to Dolphin (user id: "U071QBWBVCJ") asking if he is doing well, and wait for his reply.'
        )


if __name__ == '__main__':
    asyncio.run(main())



import os
from typing import List, overload

from litellm.types.completion import ChatCompletionMessageParam

from npiai.llm import LLM, OpenAI
from npiai.utils import logger
from npiai.types import FunctionRegistration

from npiai.core.app import App
from npiai.core.app.browser import BrowserApp
from npiai.core.base import BaseAgent
from npiai.core.hitl import HITL


class Agent(BaseAgent):
    _app: App

    def __init__(self, app: App, llm: LLM = None):
        super().__init__()
        self._app = app
        self.llm = llm or OpenAI(api_key=os.environ.get('OPENAI_API_KEY', None), model='gpt-4o')
        self.name = f'{app.name}__agent'
        self.description = app.description
        self.provider = app.provider

    def unpack_functions(self) -> List[FunctionRegistration]:
        # Wrap the chat function of this agent to FunctionRegistration
        fn_reg = FunctionRegistration(
            fn=self.chat,
            name='chat',
            description=self._app.description,
            schema={
                'type': 'object',
                'properties': {
                    'message': {
                        'type': 'string',
                        'description': f'The task you want {self._app.name} to do or the message you want to chat with {self._app.name}'
                    },
                },
                'required': ['message'],
            }
        )

        return [fn_reg]

    def use_hitl(self, hitl: HITL):
        super().use_hitl(hitl)
        self._app.use_hitl(hitl)

    async def start(self):
        await self._app.start()

    async def end(self):
        await self._app.end()

    async def chat(
            self,
            message: str,
    ) -> str:
        messages: List[ChatCompletionMessageParam] = []

        if self._app.system_prompt:
            messages.append(
                {
                    'role': 'system',
                    'content': self._app.system_prompt,
                }
            )

        messages.append(
            {
                'role': 'user',
                'content': message,
            }
        )

        return await self._call_llm(messages)

    async def _call_llm(self, messages: List[ChatCompletionMessageParam]) -> str:
        while True:
            response = await self.llm.completion(
                messages=messages,
                tools=self._app.tools,
                tool_choice='auto',
                max_tokens=4096,
            )

            response_message = response.choices[0].message

            messages.append(response_message)

            if response_message.content:
                logger.info(response_message.content)

            tool_calls = response_message.get('tool_calls', None)

            if tool_calls is None:
                return response_message.content

            results = await self._app.call(tool_calls)
            messages.extend(results)


class BrowserAgent(Agent):
    _app: BrowserApp

    def __init__(self, app: BrowserApp, llm: LLM = None):
        super().__init__(app, llm)

    async def chat(
            self,
            message: str
    ) -> str:
        if not self._app.use_screenshot:
            return await super().chat(message)

        screenshot = await self._app.get_screenshot()

        if not screenshot:
            return await super().chat(message)

        messages: List[ChatCompletionMessageParam] = []

        if self._app.system_prompt:
            messages.append(
                {
                    'role': 'system',
                    'content': self._app.system_prompt,
                }
            )

        messages.append(
            {
                'role': 'user',
                'content': [
                    {
                        'type': 'text',
                        'text': message,
                    },
                    {
                        'type': 'image_url',
                        'image_url': {
                            'url': screenshot,
                        },
                    },
                ]
            }
        )

        return await self._call_llm(messages)


@overload
def agent_wrapper(app: App, llm: LLM = None) -> Agent:
    ...


@overload
def agent_wrapper(app: BrowserApp, llm: LLM = None) -> BrowserAgent:
    ...


def agent_wrapper(app: App | BrowserApp, llm: LLM = None) -> Agent | BrowserAgent:
    if isinstance(app, App):
        return Agent(app, llm)

    if isinstance(app, BrowserApp):
        return BrowserAgent(app, llm)

    raise TypeError(f'app must be an instance of App or BrowserApp')
